﻿using System;
using System.Configuration;
using System.Linq;
using KRE.Replicate.Contracts.Data;
using KRE.Replicate.Server.Database.ConflictResolvers;
using log4net;
using Ninject;
using Raven.Abstractions.Data;
using Raven.Abstractions.Replication;
using Raven.Client;
using Raven.Client.Document;

namespace KRE.Replicate.Server.Database
{
    public class TheRaven : IRaven
    {
        private static readonly ILog Log = LogManager.GetLogger(typeof(TheRaven));

        private string _dbName;
        public string DbName => _dbName ?? (_dbName = ConfigurationManager.AppSettings.Get("RavenDbName"));
        private IDocumentStore _store;

        public TheRaven(IKernel kernel)
        {
            InitDocumentStore();
            ResetRaven();
            kernel.Bind<IDocumentStore>().ToConstant(_store);
            kernel.Bind<IDocumentSession>().ToMethod(d => _store.OpenSession(DbName));
        }

        public void ResetRaven()
        {
            RemoveOldDatabase();
            InitializeDatabase();
            Log.Debug("Reset Raven");
        }

        private void RemoveOldDatabase()
        {
            var databaseNames = _store.DatabaseCommands.GlobalAdmin.GetDatabaseNames(10);
            if (databaseNames.Contains(DbName))
            {
                DeleteDatabase(true);
            }
        }

        private void InitDocumentStore()
        {
            _store = new DocumentStore
            {
                Url = ConfigurationManager.AppSettings["RavenUrl"]
            }
                .RegisterListener(new FooResolver())
                .RegisterListener(new BarResolver()) as DocumentStore;

            if (_store == null)
            {
                throw new NullReferenceException("Documentstore is null... How did that happen?");
            }

            _store.Initialize();
        }

        private void InitializeDatabase()
        {
            _store.DatabaseCommands.GlobalAdmin.CreateDatabase(new DatabaseDocument
            {
                Id = DbName,
                Settings =
                {
                    {Constants.ActiveBundles, "Replication"},
                    {Constants.RavenDataDir, $"~\\Databases\\{DbName}"}
                }
            });
        }

        private void DeleteDatabase(bool hardDelete)
        {
            _store.DatabaseCommands.GlobalAdmin.DeleteDatabase(DbName, hardDelete);
        }

        public void StartReplication(RavenNode target)
        {
            try
            {
                AddReplicationTargetToDatabase(target);
                Log.Debug($"Replicating to {target}");
            }
            catch
            {
                Log.Error($"Failed to set up replication to {target}");
            }
        }

        public void StopReplication(RavenNode target)
        {
            try
            {
                RemoveReplicationTargetFromDatabase(target);
                Log.Debug($"No longer replicating to {target}");
            }
            catch
            {
                Log.Error($"Failed to remove replication to {target}");
            }
        }

        private void AddReplicationTargetToDatabase(RavenNode replicationTarget)
        {
            using (var session = _store.OpenSession(DbName))
            {
                var replicationDocument = session.Load<ReplicationDocument>(Constants.RavenReplicationDestinations);
                if (replicationDocument == null)
                {
                    replicationDocument = new ReplicationDocument();
                    session.Store(replicationDocument, Constants.RavenReplicationDestinations);
                }

                replicationDocument.Destinations.Add(new ReplicationDestination
                {
                    Url = replicationTarget.Url,
                    Database = replicationTarget.Name
                });

                session.SaveChanges();
            }
        }

        private void RemoveReplicationTargetFromDatabase(RavenNode replicationTarget)
        {
            using (var session = _store.OpenSession(DbName))
            {
                var replicationDocument = session.Load<ReplicationDocument>(Constants.RavenReplicationDestinations);
                if (replicationDocument == null)
                {
                    return;
                }

                replicationDocument.Destinations.RemoveAll(rd => rd.Database == replicationTarget.Name);

                session.SaveChanges();
            }
        }

        public void DisableDatabase()
        {
            _store.DatabaseCommands.GlobalAdmin.DeleteDatabase(DbName);
        }

        public void EnableDatabase()
        {
            InitializeDatabase();
        }
    }
}
